"""
    db.py
    DB Class

    ...

    json types are possible as inputs, just pass a python dict as field value

"""
import asyncio
import asyncpg
import datetime
import re
from pathlib import Path
import orjson
import numpy as np
from dl2050utils.core import listify, oget, check_str, check_dict, check
from dl2050utils.log import BaseLog
from dl2050utils.fs import pickle_save, pickle_load

# TODO
#  1) db should Raise Error on insert/update/delete failures
#  2) SELECT: nrows: returning 2 for results with one row, and 1 instead of zero when there are no results

# ####################################################################################################
# Helper functions
# ####################################################################################################

def strip(e):
    if type(e)!=str: return e
    e = e.replace('\'', '')
    e.replace('\"', '')
    e.replace('\n', ' ')
    return e

def get_repr(e):
    if e is None: return 'null'
    if type(e)==str: return f"'{strip(e)}'"
    if type(e) in [int,float]: return str(e)
    if type(e)==asyncpg.pgproto.pgproto.UUID: return f"'{e}'"
    if type(e)==datetime.datetime:
        s = e.strftime("%Y-%m-%d %H:%M:%S")
        return f"'{s}'"
    if type(e)==np.str_: return f"'{strip(str(e))}'"
    if np.issubdtype(type(e), np.str_): return f"'{strip(str(e))}'"
    if np.issubdtype(type(e), np.integer) or np.issubdtype(type(e), np.floating): return str(e)
    if type(e)==list:
        items = [f'"{str(e1)}"' for e1 in e]
        return f"'{{{' ,'.join(items)}}}'"
    if type(e)==dict:
        s = orjson.dumps(e, option=orjson.OPT_SERIALIZE_NUMPY).decode()
        return f"'{s}'"
    return str(e)

def fix_types(d):
    for k in d.keys():
        if isinstance(d[k], str):
            d[k] = d[k].strip()
        if isinstance(d[k], datetime.datetime):
            d[k] = d[k].strftime("%Y-%m-%d %H:%M:%S")
        if isinstance(d[k], datetime.date):
            d[k] = d[k].strftime("%Y-%m-%d")
    return d

def parse_filters(fs, tbl):
    """
    Filters can be either a list of dicts or a dict.
    When a dict, all attrs/values represent key,val conditions. dict filter is converted in list filters.
    List filter are lists with dict with one filter condition each, with col,val attrs, and optionally the op attr.
    The op attr can be =,<,>,<=,>=,!=
    """
    if fs is None or type(fs)!=dict and type(fs)!=list: return []
    if type(fs)==dict:
        fs = [{'col':e, 'val':fs[e]} for e in fs]
    if type(fs)==list:
        fs2 = []
        for e in fs:
            f = check_dict(e, keys=['col','val','op'])
            if f is not None and 'col' in f and 'val' in f:
                f2 = {'col':f"{tbl}.{f['col']}", 'val':f['val'], 'op':oget(f,['op'],'=')}
                fs2.append(f2)
        return fs2
    if type(fs)==dict:
        return [{'col':check_str(e,n=1024), 'val':check(fs[e]), 'op':'='} for e in fs]
    return []

def parse_join(join):
    """
    Used to filter the results based on a condition col=val in a join tbl.
    join is a dict with attrs:
        tbl2: the join tbl
        key1: the master tbl join col
        key2: the join tbl key
        col: col where condition is testes
        val: val of the condition
    Example:
        {'tbl2':'orgs', 'key1':'org', 'key2':'id', 'col':'name', 'val':'MHR'}
    """
    cols = ['tbl2', 'key1', 'key2', 'col', 'val']
    if type(join)!=dict: return None
    join2 = {**join}
    for e in cols:
        if e not in join2: return None
    return {e:check(join2[e]) for e in cols}

def parse_lookups(lookups):
    """
    Used to join with lookup tbls on lookup cols and extract lookup vals.
    lookups is a list of dicts with attrs:
        col: the lookup col from the master tbl
        tbl: the lookup tbl
        key: the key col of the lookup tbl, optional, defaults to id
        val: the val col of the lookup tbl, optional, defaults to name
    Example:
        [{'col':'org', 'tbl':'orgs', 'key':'id', 'val':'name',}]
    """
    if type(lookups)!=list or not len(lookups): return []
    cols = ['col', 'tbl', 'key', 'val']
    lookups2 = []
    for l in lookups:
        if 'col' not in l or 'tbl'  not in l: continue
        l['key'] = oget(l,['key'],'id')
        l['val'] = oget(l,['val'],'name')
        l2 = {e:check(l[e]) for e in cols}
        lookups2.append(l2)
    return lookups2

def get_select_q(tbl, join=None, lookups=None, filters=None, sfilters=None, cols=None, sort=None, ascending=True,
                 offset=None, limit=None):
    qcols = f'{tbl}.*' if cols is None else ', '.join(cols)
    # Parse filters and sfilters
    fs,sfs = parse_filters(filters, tbl),parse_filters(sfilters, tbl)
    # Insert filter by join condition
    qjoins = ''
    join = parse_join(join)
    if join is not None:
        qjoins += f" JOIN {join['tbl2']} ON {tbl}.{join['key1']}={join['tbl2']}.{join['key2']}"
        fs.append({'col':f"{join['tbl2']}.{join['col']}", 'val':join['val'], 'op':'='})
    # Insert lookups
    lookups = parse_lookups(lookups)
    for l in lookups:
        qcols += f", {tbl}.{l['col']} as {l['col']}_id, {l['tbl']}.{l['val']} as {l['col']}_name"
        # qjoins += ' JOIN '
        qjoins += ' LEFT OUTER JOIN '
        qjoins += f"{l['tbl']} ON {tbl}.{l['col']}={l['tbl']}.{l['key']} "
    # Prepare full filters and sfilters
    qwhere = ''
    if len(fs):
        qcond = ' AND '.join([f"{e['col']}{e['op']}{get_repr(e['val'])}" for e in fs])
        qwhere = f" WHERE {qcond}"
    if len(sfs):
        qwhere += ' WHERE ' if not len(fs) else ' AND '
        qwhere += ' AND '.join([f"{e['col']} ILIKE '%{e['val']}%'" for e in sfs])
    # Insert order
    qorder = ''
    if isinstance(sort, str):
        qorder = f" ORDER BY {sort} " + ("ASC" if ascending else "DESC")
    # Insert offset and limit
    qpag = ''
    offset,limit = offset or 0,limit or 32
    if offset is not None: qpag += f" OFFSET {offset}"
    if limit is not None: qpag += f" LIMIT {limit}"
    # Prepare query and replace every '=null' by ' is null'
    q = f"SELECT {qcols} FROM {tbl}{qjoins}{qwhere}{qorder}{qpag};"
    q = re.sub(r'=null', ' is null', q, flags=re.IGNORECASE)
    return q

def prep_query(q):
    """Strips q and adds LIMIT clause if it is a SELECT and it is not present."""
    q = q.strip()
    if q[:6].upper()!='SELECT': return q
    has_semicolon = q.endswith(';')
    # If there's a semicolon, remove it temporarily
    if has_semicolon: q = q[:-1].strip()
    # Check if the query already contains 'LIMIT <int>' and append if not
    if not re.search(r'LIMIT \d+$', q, re.IGNORECASE): q += " LIMIT 32"
    # Re-append the semicolon if it was originally there
    if has_semicolon: q += ';'
    return q

async def calc_query_rows(con, q):
    """
        Returns the exepected number of rows in a query.
        The number of rows is obtained by changing the query to count(*) and dropping the sort.
        TODO: Implement Explain/Query Plan approach, currently faliling on joins
    """
    q1 = re.sub(r'SELECT .*? FROM', 'SELECT count(*) FROM', q, flags=re.IGNORECASE)
    q1 = re.sub(r'ORDER BY .*?$', '', q1, flags=re.IGNORECASE)
    q1 = re.sub(r'OFFSET \d+', '', q1, flags=re.IGNORECASE)
    q1 = re.sub(r'LIMIT \d+', '', q1, flags=re.IGNORECASE)
    res = await con.fetchrow(q1)
    if res is None: return None
    res = dict(res)
    return oget(res,['count'])
    # db.sync_select('pg_class', cols=['reltuples'], filters={'relname': 'diagsg'})
    # res = await con.fetchrow(f'explain(format json) {q}')
    # if res is None: return None
    # res = json.loads(res['QUERY PLAN'])
    # if res is None or not len(res): return None
    # return oget(res[0],['Plan','Plan Rows'])

# ####################################################################################################
# DB Class
#
#   startup
#   shutdown
#
#   query
#   select
#   select_one
#   insert
#   update
#   update_or_insert
#   delete
#   get_seq
#
# ####################################################################################################

class DB():
    """
        Class for managing interactions with a PostgreSQL database.

        Constructor Parameters:
                - cfg (dict, optional): DB connection parameters (default: None, and a basic connection is provided).
                - log (BaseLog, optional): Logging mechanism for error and status messages (default: None).
                - dbname (str, optional): Name of the database to connect to (default: None).

            Methods:
                - async startup(min_size=5, max_size=20, loop=None)
                  Initiates the database connection pool with optional size and loop parameters.
                - shutdown()
                  Gracefully closes the database connection pool.
                - async execute(q)
                  Executes SQL queries (non-SELECT) and returns the result.
                - async def query(q, one=False, nrows=False, offset=None, limit=None)
                  Executes SELECT queries and retrieves data.
                - async get_trows(tbl) (Internal)
                  Retrieves the total number of rows that will result in a querty (through query planner).
                - async select(tbl, join=None, filters=None, sfilters=None, cols='*', sort=None, ascending=True, offset=None, limit=None, one=False)
                  Retrieves data from a table with filtering, sorting, and pagination options.
                - async select_one(tbl, filters=[])
                  Retrieves a single row from a table with optional filters.
                - async insert(tbl, d, return_key=None)
                  Inserts data into a table and optionally returns the inserted key.
                - async update(tbl, ks, d)
                  Updates rows in a table based on provided keys.
                - async update_or_insert(tbl, ks, d)
                  Updates rows if they exist, otherwise inserts new rows.
                - async delete(tbl, k, v)
                  Deletes rows from a table based on a key-value pair.
                - update_stats()
                  Performs a vacuum analyze operation for database statistics update (needed for query planner to work properly)

                The DB class assumes the use of a PostgreSQL database and proper configuration parameters for connection.
                Error messages and context information are logged for debugging purposes, enhancing reliability and maintainability.
                Ensures data integrity and security by validating and sanitizing input data.

                The filtering in define by two optional distint ways:
                    1) As an array of dict objects of the for {'col':col, 'val':val, 'op':op.} Op can be wither =,>,>=,<,<=,!=. 
                    2) Single dict with all the keys/vals.
    """
    def __init__(self, cfg=None, log=None, dbname=None, passwd=None):
        if cfg is None: cfg={'db':{'host':'db','port':5432,'user':'postgres','passwd':'rootroot','dbname':'postgres'}}
        self.cfg, self.LOG = cfg, log or BaseLog()
        host = oget(cfg,['db','host'],'db')
        port = oget(cfg,['db','port'],5432)
        user = oget(cfg,['db','user'],'postgres')
        passwd = oget(cfg,['db','passwd'],passwd)
        if dbname is None:
            dbname = oget(cfg,['db','dbname'],'postgres')
        self.url = f'postgres://{user}:{passwd}@{host}:{port}/{dbname}'
        self.dbname = dbname

    async def startup(self, min_size=5, max_size=20, loop=None):
        try:
            self.pool = await asyncpg.create_pool(self.url, min_size=min_size, max_size=max_size, loop=loop)
        except Exception as e:
            self.LOG(4, 0, label='DBPG', label2='STARTUP', msg=str(e))
            return True
        self.LOG(2, 0, label='DBPG', label2='STARTUP', msg=f'CONNECTED POOL to {self.dbname}')
        return False

    def shutdown(self):
        self.pool.terminate()
        self.LOG(2, 0, label='DBPG', label2='shutdown', msg='DISCONNECTED')
        return False
    
    async def execute(self, q):
        con = await self.pool.acquire()
        try:
            res = await con.execute(q)
        except Exception as err:
            self.LOG(4, 0, label='DBPG', label2='execute', msg={'error_msg': str(err), 'query': q})
            return None
        finally:
            await self.pool.release(con)
        return res

    async def query(self, q, one=False, nrows=False):
        con = await self.pool.acquire()
        q = prep_query(q)
        nr = None
        if q[:6].upper()!='SELECT':
            one = True # accounts for returning in insert
        if one:
            nrows = False
        try:
            if(one):
                res = await con.fetchrow(q)
                if res is None:
                    await self.pool.release(con)
                    return None
            else:
                if nrows:
                    nr = await calc_query_rows(con, q)
                    if nr is None:
                        self.LOG(4, 0, label='DB', label2='calc_query_rows', msg=q)
                        await self.pool.release(con)
                        return None
                res = await con.fetch(q)
        except Exception as e:
            self.LOG(4, 0, label='DB', label2='query', msg=str(e))
            return None
        finally:
            await self.pool.release(con)
        if res is None: return None
        if(one): return fix_types(dict(res))
        res = [fix_types(dict(row)) for row in res]
        if nrows: return {'data':res, 'nrows':nr}
        return res
    
    async def get_trows(self, tbl):
        q = f"select reltuples as nrows from pg_class where relname='{tbl}'"
        con = await self.pool.acquire()
        res = await con.fetchrow(q)
        if res is None:
            return -1
        return int(dict(res)['nrows'])

    async def select(self, tbl, lookups=None, join=None, filters=None, sfilters=None, cols='*', sort=None, ascending=True,
                     offset=None, limit=None, one=False):
        q = get_select_q(tbl, lookups=lookups, join=join, filters=filters, sfilters=sfilters, cols=cols, sort=sort, ascending=ascending,
                         offset=offset, limit=limit)
        return await self.query(q, one=one, nrows=True)

    async def select_one(self, tbl, filters=[]):
        return await self.select(tbl, filters=filters, one=True)
    
    async def insert(self, tbl, d, return_key=None):
        q = f"INSERT INTO {tbl} ("
        for k in d.keys(): q += f"{k}, "
        q = q[:-2] + ") VALUES ("
        for k in d.keys(): q = q + get_repr(d[k]) + ', '
        q = q[:-2] + ")"
        if return_key is not None:
            q += f" returning {return_key};"
            res = await self.query(q, one=True)
            if res is not None and len(res):
                return res[return_key]
            return None
        res = await self.execute(q)
        if res=='INSERT 0 1':
            return False
        return True

    async def update(self, tbl, ks, d):
        ks = listify(ks)
        for k in ks:
            if not k in d:
                self.LOG(4, 0, label='DBPG', label2='update', msg=f'key error: {k}')
                return True
        setvars = ', '.join([f"{k}={get_repr(d[k])}" for k in d if k not in ks+['nrows']])
        q = f"UPDATE {tbl} SET {setvars}"
        filters = ' AND '.join([f"{k}='{d[k]}'" for k in ks if d[k] is not None])
        q += f" WHERE {filters};"
        res = await self.execute(q)
        if res is None:
            return True
        n = int(res[7:])
        if n==0:
            return True
        return False

    async def update_or_insert(self, tbl, ks, d):
        ks = listify(ks)
        filters = {e:d[e] for e in ks}
        row = await self.select_one(tbl, filters=filters)
        if row is None:
            return await self.insert(tbl, d)
        return await self.update(tbl, ks, d)

    # Missing multi key delete
    async def delete(self, tbl, k, v):
        if k is None or v is None: return True
        q = f"DELETE FROM {tbl} WHERE {k}='{v}'"
        res = await self.execute(q)
        if res is None: return True
        n = int(res[7:])
        if n==0: return True
        return False
    
    async def get_seq(self, seqname):
        """ Return the next sequence value of the sequence seqname. """
        res = await self.query(f"select nextval('{seqname}')")
        return oget(res,[0,'nextval'])
    
    def set_seq(self, seqname):
        pass

    def update_stats(self):
        self.sync_query('vacuum analyze')

    def sync_startup(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.startup(*args, **kwargs))
    def sync_query(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.query(*args, **kwargs))
    def sync_select(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.select(*args, **kwargs))
    def sync_select_one(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.select_one(*args, **kwargs))
    def sync_insert(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.insert(*args, **kwargs))
    def sync_update(self,  *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.update(*args, **kwargs))
    def sync_update_or_insert(self,  *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.update_or_insert(*args, **kwargs))
    def sync_delete(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.delete(*args, **kwargs))
    def sync_get_seq(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.get_seq(*args, **kwargs))

# ####################################################################################################
# Utility DB functions
# ####################################################################################################

def row_to_rec(row, lcols):
    """
    Adapts DB output to for front-end consuption.
    Row is produced in the DB Class and has _id and _name suffixes on lookups.
    Rec converts the suffix into an object to be sent to the front-end.
    row_to_rec converts a Row dict into a Rec dict, colapsing the id and name attrs into dicts with id,name attrs,
    for all cols with prefix incuded in lcols.

    Example for lcols=['org']:
        {'id':247, 'org_id':3215,'org_name':'RP'} -> {'id':247, 'org':{'id': 3215, 'name': 'RP'}}
    """
    exclude = [f'{e}_id' for e in lcols] + [f'{e}_name' for e in lcols]
    rec = {k:row[k] for k in row if k not in exclude}
    for e in lcols:
        rec[e] = {'id':row[f'{e}_id'], 'name':row[f'{e}_name']}
    return rec

def res_to_recs(res, lcols):
    """Applies row_to_rec to a list of res."""
    if type(lcols)!=list: return res
    if type(res)==dict:
        if not 'data' in res: return None
        res = res['data']
    if type(res)!=list: return None
    return [row_to_rec(e, lcols) for e in res]

def rec_to_row(rec):
    """
    Converts Front-end Rows with object columns with id,name attrs by colapsing all the lookup dicts into the id attr.
    Example:
        {'id':247, 'org': {'id':3215, 'name':'RP'}} -> {'id':247, 'org':3215}
    """
    return {k:rec[k]['id'] if type(rec[k])==dict else rec[k] for k in rec}

def run_slq_cmds(db, q):
    for q in q.split(';'):
        if len(q)<10: continue
        db.sync_query(q+';')

def run_sql_file(db, fname):
    f = open(fname, 'r')
    qs = f.read()
    for q in [q for q in qs.split(";")]:
        db.sync_query(q)

def db_insert_df(db, tbl, df, dmap=None, delete=False):
    if delete:
        db.sync_query(f'delete from {tbl}')
    if dmap is None:
        dmap={c:c for c in df.columns}
    else:
        df = df.rename(columns=dmap)
    cols = [dmap[c] for c in dmap]
    df = df[cols]
    for i in range(len(df)):
        row = df.iloc[i]
        d = {c:row[c] for c in cols}
        if db.sync_insert(tbl, d):
            print('DB insert error')
    return False

def log_and_return(msg):
    print(msg)
    return msg

def db_export_tbl(db, p, tbl):
    res = db.sync_select(tbl, limit=int(1e9))
    if res is None or not 'data' in res:
        return 1
    return pickle_save(p, res['data'])

def db_import_tbl(db, p, tbl, delete=False):
    p = Path(p)
    if not p.with_suffix('.pickle').is_file():
        return log_and_return(f'Error importing {tbl}: file {p} not found')
    rows = pickle_load(p)
    if rows is None:
        return log_and_return(f'Cant read {p}')
    if delete:
        if db.sync_query(f'delete from {tbl}'):
            return log_and_return(f'Error deleting tbl {tbl}')
    n = 0
    for row in rows:
        res = db.sync_insert(tbl, row)
        if res:
            return log_and_return(f'Error inserting record into {tbl}: {row}')
        n += 1
    print(f'Imported {n} records into {tbl}')
    return 0

def db_disable_serial(db, tbl, col):
    db.sync_query(f"alter table {tbl} alter column {col} drop default")

def db_enable_serial(db, tbl, col):
    n = db.sync_query(f"select max({col}) from {tbl}")[0]['max']
    db.sync_query(f"select pg_catalog.setval('public.{tbl}_{col}_seq', {n}, true)")
    db.sync_query(f"alter table {tbl} alter column {col} set default nextval('{tbl}_{col}_seq')")



import hashlib
import pathlib
from pathlib import Path
import numpy as np
from dl2050utils.common import pickle_save, pickle_load

class FileDB():
    """
    A class for managing file-based storage using a structured directory hierarchy,
    with an optional hashing mechanism for non-integer keys.
    Attributes:
        base_path (Path): The root directory for storing files.
        max_value (int): The maximum range for hashed values to avoid excessive directory depth.
    Methods:
        get_dir(key): Generates a directory path based on the hash or integer representation of the key.
        get_name(key, pre='', ext=''): Generates a file name based on the key, with optional prefix and extension.
        save(key, data, pre='', ext='', save_f=pickle_save): Saves data to a file in the appropriate directory.
        load(key, pre='', ext='', load_f=pickle_load): Loads data from a file based on the key, prefix, and extension.
    """
    def __init__(self, base_path, max_value=10**12):
        """
        Initializes FileDB with a specified base path and maximum hash value.
        Args:
            base_path (str or Path): The root directory where files will be stored.
            max_value (int, optional): Limits the hash range for directory distribution. Defaults to 10^12.
        """
        self.base_path,self.max_value = Path(base_path),max_value
        self.base_path.mkdir(parents=True, exist_ok=True)
    def get_dir(self, key):
        """
        Generates a nested directory path for a given key.
        Args:
            key (int or str): The unique identifier for the file. If a string, it is hashed using MD5.
        Returns:
            Path: A directory path within the base path, structured in 3-levels based on the key hash.
        """
        if type(key)==int:
            seq = key
        else:
            key = str(key)
            hash_object = hashlib.md5(key.encode('utf-8'))
            seq = int(hash_object.hexdigest(), 16)
        seq = seq % self.max_value
        s = f'{seq:012}'
        dir_path = Path(f'{self.base_path}/{s[:3]}/{s[3:6]}/{s[6:9]}')
        return dir_path
    def get_name(self, key, pre='', ext=''):
        """
        Generates a file name using the key, with optional prefix and extension.
        Args:
            key (int or str): The unique identifier for the file.
            pre (str, optional): A prefix for the file name.
            ext (str, optional): A file extension.
        Returns:
            str: The generated file name.
        """
        return f'{pre}{key}{ext}'
    def save(self, key, data, pre='', ext='', save_f=pickle_save):
        """
        Saves data to a file in the directory corresponding to the key.
        Args:
            key (int or str): The unique identifier for the file.
            data (any): The data to be saved.
            pre (str, optional): A prefix for the file name.
            ext (str, optional): The file extension. Use '.npy' for NumPy arrays.
            save_f (callable, optional): The save function to use (defaults to pickle_save).
        Returns the return of the called save function.
        """
        dir_path = Path(self.get_dir(key))
        dir_path.mkdir(parents=True, exist_ok=True)
        file_name = self.get_name(key, pre=pre, ext=ext)
        p = dir_path/file_name
        if ext=='.npy': return np.save(p, data)
        return save_f(p, data)
    def load(self, key, pre='', ext='', load_f=pickle_load):
        """
        Loads data from a file based on the key, prefix, and extension.
        Args:
            key (int or str): The unique identifier for the file.
            pre (str, optional): A prefix for the file name.
            ext (str, optional): The file extension. Use '.npy' for NumPy arrays.
            load_f (callable, optional): The load function to use (defaults to pickle_load).
        Returns:
            any: The loaded data.
        """
        dir_path = Path(self.get_dir(key))
        file_name = self.get_name(key, pre=pre, ext=ext)
        p = dir_path/file_name
        if ext=='.npy': return np.load(p)
        return load_f(p)



"""
restapp.py
App Class

Defines the logic for the REST App, including startup, shudown, Auth routes, downstream services routes and Custom Application routes.

The Auth routes are defined in the Auth module.

The downstream services routes include:
    /api/contact
    /api/get_meta
    /api/download_req
    /api/upload_req
    /api/download_workfile/{fname}
    /api/upload_workfile
    /api/publish_job
    /api/get_jobs
    /api/get_job

The Custom Application are defined in the routes dict with the following structure:
    routes = {
        <route_path_string>: <route_defenition>,
        ...
    }
    <route_defenition> is a dict with keys 'f','args' and 'kwargs', with the function name implementing the route along with
    corresponding args and kwargs as lists of strings.

    For example:
        routes = {
            '/api/health': {'f': health, 'args':[], 'kwargs':[]},
            ...
        }

Every custom function implementing the route must observe the following:
    1) All inputs parsed from the Request as function parametes that must be defined in the lists args and kwargs above;
    2) All returns should be a dict with the format {status:<status>, ...}:
        - <status> should be either 'OK' or 'ERROR'
        - If 'status':'OK' extra attributes may be present and will be sent to the client
        - If 'status':'ERROR', attribute 'error_msg' should be present, optionally along with 'error_code'
        - If status is not present, 'status':'OK' is automatically inserted;
        - If the return is not a dict, a dict is created with {'status':'OK', 'result':<return>}
        - If None is returned, a dict is created with only {'status':'OK'}
        Examples:
            return {'status':'ERROR', 'error_msg':'Error message', 'error_code':'Error code'} # Client gets what is returned
            return {'data':[arg1, kwarg1]} # Client gets {'status':'OK', 'data':[arg1, kwarg1]}
            return [1,2,3] # Client gets {'status':'OK', 'result':[1, 2, 3]}
            return None # Client gets {'status':'OK'}
            return {'status':'ERROR', 'error_msg':'Error message', 'error_code':'Error code'} # Client gets what is returned
    3) All normal returns are converted into HTTP responses with HTTP status 200 and the return data as JSON
    4) It is possible to raise an exception through log_and_raise_rest(msg), that is converted into an HTTP response
       with HTTP status defaulting to 500 and the error data
    5) If the function has an uncatched exception (bug), the App catches it and converts in the following response:
       {'status':'ERROR', 'error_type':'EXCEPTION', 'error_msg': <exception_details>} and then sent as an HTTP response
       with HTTP status 400 and the error data as JSON

Regarding files upload/downloand, consider the following:
    1) The routes /api/download_req and /api/upload_req provice signed URLs for downnload/upload.
       The signed URLs can be obtained from the File Server or Google Cloud Storage.
       They are intended to offload the REST server of dealing with files, proxying to external services.
       The user namespace is limited to <bucket>.
    2) The routes /api/download_workfile and /api/upload_workfile provides downnload/upload functionality for
       registered users.
       User have isolated namespace under the workfiles namespace.

The remaing methods deal with jobs lifecycle and there is a method to register contact requests.

"""
from datetime import datetime, timezone
from operator import itemgetter
from pathlib import Path
from starlette.responses import FileResponse
from starlette.routing import Route
from starlette.authentication import requires
from dl2050utils.core import oget, LexIdx, get_uid
from dl2050utils.gs import GS
from dl2050utils.restutils import HTTPException, log_and_raise_exception, log_and_raise_rest, log_and_raise_service
from dl2050utils.restutils import rest_ok,enforce_required_args, get_optional_args, form_files_upload
from dl2050utils.restutils import fs_download_signed_url, fs_upload_signed_url, gcs_download_signed_url, gcs_upload_signed_url

class App():

    def __init__(self, cfg, LOG, NOTIFY, db, mq, auth, path, routes=[], appstartup=None, perm=None):
        self.service = cfg['service']
        self.path = path
        self.cfg,self.LOG,self.NOTIFY,self.db,self.mq,self.auth = cfg,LOG,NOTIFY,db,mq,auth
        self.routes,self.appstartup,self.perm = routes,appstartup,perm
        self.d = {'cfg':cfg, 'LOG':LOG, 'path':path, 'db':db, 'mq':mq}
        self.fs = GS(self.service) if oget(cfg,['gcloud','gs_key']) is not None else None
        # self.fs_url,self.fs_secret = oget(self.cfg,['fs','url']),oget(self.cfg,['fs','secret'])

    # ####################################################################################################
    # Helper methods
    # ####################################################################################################

    async def db_increase(self, tbl, k, kv, c, prefix=''):
        row = await self.db.select_one(tbl, {k:kv})
        if row is None or len(row)==0:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB increase error: select')
        row[c] += row[c]+1
        res = await self.db.update(tbl, k, row)
        if res:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB increase error: update')

    # ####################################################################################################
    # Internal methods: startup, shutdown, get_routes
    # ####################################################################################################

    async def startup(self):
        model,meta = oget(self.cfg,['app','model']),None
        # if model is not None: meta = await get_meta(self.path, self.db, model)
        # self.d['model'] = model
        # self.d['meta'] = meta
        self.d['meta'] = None
        if self.appstartup is None: return False
        return await self.appstartup(self.d)

    def shutdown(self):
        self.LOG(2, 0, label='APP', label2='shutdown', msg='OK')
        return False   

    def get_routes(self):
        BASE_ROUTES = [
            Route('/api/contact', endpoint=self.contact, methods=['POST']),
            Route('/api/get_meta', endpoint=self.get_meta, methods=['GET']),
            Route('/api/get_signed_urls', endpoint=self.get_signed_urls, methods=['POST']),
            Route('/api/download_req', endpoint=self.download_req, methods=['POST']),
            Route('/api/upload_req', endpoint=self.upload_req, methods=['POST']),
            Route('/api/download_workfile/{fname:path}', endpoint=self.download_workfile, methods=['GET']),
            Route('/api/upload_workfile/{fname:path}', endpoint=self.upload_workfile, methods=['POST']),
            Route('/api/publish_job', endpoint=self.publish_job, methods=['POST']),
            Route('/api/get_jobs', endpoint=self.get_jobs, methods=['POST']),
            Route('/api/get_job', endpoint=self.get_job, methods=['POST']),
            Route('/api/ulist_get', endpoint=self.ulist_get, methods=['POST']),
            Route('/api/ulist_insert', endpoint=self.ulist_insert, methods=['POST']),
            Route('/api/ulist_update', endpoint=self.ulist_update, methods=['POST']),
            Route('/api/ulist_delete', endpoint=self.ulist_update, methods=['POST']),
        ]
        APP_ROUTES = [Route(e, endpoint=self.app_route, methods=['POST']) for e in self.routes]
        return BASE_ROUTES + APP_ROUTES
    
    # ####################################################################################################
    # REST methods for Apps
    # ####################################################################################################

    async def contact(self, request):
        """
        """
        data = await request.json()
        d = enforce_required_args(self.LOG, data, ['email','name','msg'], label='AUTH', label2='check_user')
        d['ts'] = datetime.now(timezone.utc)
        err = await self.db.insert('contacts', d)
        if err:
            log_and_raise_service(self.LOG, label='REST', label2='contact', msg='DB access error')
        html = '<h1>Name</h1><p>{d["email"]</p><h1>Email</h1><p>{d["email"]</p><h1>Message</h1><p>{d["msg"]</p>}'
        self.notify.send_mail_async(d['email'], subject='Contact from Website (cardiolife.global)', html=html)
        return rest_ok()

    @requires('authenticated')
    async def get_meta(self, request):
        """Returns the meta information."""
        return rest_ok(self.d['meta'])
    
    @requires('authenticated')
    async def get_signed_urls(self, request):
        """
        Provides signed urls for every path defined in the files parameter.
        Keep the files tree structure on the File Server.
        All paths will have the same uid prefix. A new uid is allways generated on this request.
        Returns dict with two lists, the upload_urls and download_urls.
        """
        if self.fs is None:
            log_and_raise_service(self.LOG, label='REST', label2='download_req', msg='Signed urls not available')
        data = await request.json()
        # files = get_param(request, 'files', list, data=data, LOG=self.LOG)
        args = enforce_required_args(self.LOG, data, ['files'], label='REST', label2='get_signed_urls')
        files = args['files']
        uid = get_uid()
        upload_urls,download_urls = [],[]
        bucket = f'{self.service}-apiserver'
        for file in files:
            upload_url,download_url = self.fs.urls(bucket, f'{uid}/{file}')
            upload_urls.append(upload_url),download_urls.append(download_url)
        return rest_ok({'upload_urls':upload_urls, 'download_urls':download_urls})
    
    @requires('authenticated')
    async def upload_req(self, request):
        """
            Returns an upload signed URL for <bucket>/<key> for the specified key.
            The user namespace is limited to <bucket>.
            The URL will be executed either by the File Server or the Cloud File Server (TO BE DEFINE HOW)
            A 24h default timeout is set.
            TODO: consider prefix <email> or <org>
            TODO: Control size of user uploads
            TODO: Resumable uploads with GCE
        """
        pass
        # if not self.fs_url or not self.fs_secret:
        #     log_and_raise_service(self.LOG, label='REST', label2='download_req', msg='File server not defined')
        # u = await self.auth.check_auth(request)
        # uid = u['uid']
        # email = await self.auth.get_email_from_uid(uid)
        # await self.db_increase('users', 'id', u['uid'], 'uploads', 'uploads')
        # data = await request.json()
        # args = ['bucket','key','size']
        # args = enforce_required_args(self.LOG, data, args, label='REST', label2='upload_req')
        # self.perm(self.d, u, request.url.path, args)
        # # Key is composed only by <key> for now, consideer <email><key> or <org><key> later
        # # key = f"{email}/{args['key']}"
        # key = args['key']
        # timeout = 7*24*3600
        # if self.fs_url!='gcs':
        #     baseurl = f'https://{self.fs_url}'
        #     url = fs_upload_signed_url(self.fs_secret, baseurl, args['bucket'], key, args['size'], timeout=timeout)
        #     down_load_url = fs_download_signed_url(self.fs_secret, baseurl, args['bucket'], key, timeout=timeout)
        # else:
        #     url = gcs_upload_signed_url(args['bucket'], args['key'], timeout=timeout, size=args['size'])
        #     down_load_url = gcs_download_signed_url(args['bucket'], key, timeout=timeout)
        # self.LOG(2, 0, label='APP', label2='upload_req', msg={'url':url})
        # return rest_ok({'url':url, 'down_load_url':down_load_url, 'info':{'Host':oget(request.headers,['Host'])}})
    
    @requires('authenticated')
    async def download_req(self, request):
        """
            Returns a download signed URL for <bucket>/<key> for the specified key.
            The user namespace is limited to <bucket>/<email> and <email> is prepended to <key> as dir.
            The URL will be executed either by the File Server or the Cloud File Server (TO BE DEFINE HOW)
            A 24h default timeout is set.
            TODO: consider prefix <email> or <org>
        """
        pass
        # u = await self.auth.check_auth(request)
        # uid = u['uid']
        # email = await self.auth.get_email_from_uid(uid)
        # await self.db_increase('users', 'id', u['uid'], 'downloads', 'downloads')
        # data = await request.json()
        # args = ['bucket','key']
        # args = enforce_required_args(self.LOG, data, args, label='REST', label2='download_req')
        # self.perm(self.d, u, request.url.path, args)
        # # Key is composed only by <key> for now, consideer <email><key> or <org><key> later
        # # key = f"{email}/{args['key']}"
        # key = args['key']
        # timeout = 7*24*3600
        # if self.fs_url!='gcs':
        #     baseurl = f'https://{self.fs_url}'
        #     url = fs_download_signed_url(self.fs_secret, baseurl, args['bucket'], key, timeout=timeout)
        # else:
        #     url = gcs_download_signed_url(args['bucket'], key, timeout=timeout)
        # self.LOG(2, 0, label='APP', label2='download_req', msg={'url':url})
        # return rest_ok({'url':url})

    @requires('authenticated')
    async def download_workfile(self, request):
        """
            Downloads a workfile file fname belonging to a registered user.
            Workfiles are usually results produced by workers (throught MQ).
            Workfiles are stored in the folder '$DATAPATH/workfiles/<service>.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        fname = request.path_params['fname']
        p = Path(f'/data/{self.service}/workfiles/{fname}')
        if not p.is_file():
            log_and_raise_service(self.LOG, label='REST', label2='download_workfile', msg=f'File not found: {str(p)}')
        return FileResponse(str(p), media_type='application/vnd.ms-excel', filename=str(p))

    @requires('authenticated')
    async def upload_workfile(self, request):
        """
            Uploads a workfile file fname belonging to a registered user.
            Workfiles for upload are usually content produced by the users.
            Workfiles are stored in the folder '$DATAPATH/workfiles/<service>'.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        fname = request.path_params['fname']
        p = Path(f'/data/{self.service}/workfiles/{fname}')
        max_size = int(1e9)
        form = await request.form()
        if form is None:
            log_and_raise_service(self.LOG, label='REST', label2='upload_workfile', msg=f'Upload form error')
        await form_files_upload(self.LOG, form.items(), p, max_size=max_size)
        return rest_ok()

    @requires('authenticated')
    async def publish_job(self, request):
        """
            Publishes a new job, calling mq.publish after checking permissions.
            Jobs require the user email as an attribute.
        """
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['q','payload'], label='REST', label2='publish_job')
        q,payload = args['q'],args['payload']
        self.perm(self.d, u, request.url.path, args)
        jid = await self.mq.publish(q, email, payload)
        if jid is None :
            log_and_raise_service(self.LOG, label='REST', label2='publish_job', msg='MQ publish error')
        self.LOG(2, 0, label='APP', label2='publish_job')
        return rest_ok({'jid': jid})

    @requires('authenticated')
    async def get_jobs(self, request):
        """Gets all jobs from a user, optionaly filtered by qname, pending or jobs not done."""
        u = await self.auth.check_auth(request)
        uid = u['uid']
        email = await self.auth.get_email_from_uid(uid)
        data = await request.json()
        kwargs = get_optional_args(data, ['qname','pending','not_done'])
        jobs = await self.mq.get_jobs(email, **kwargs)
        return rest_ok(jobs)

    @requires('authenticated')
    async def get_job(self, request):
        """Gets a job details, identified by jid."""
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['jid'], label='REST', label2='get_job')
        job = await self.mq.get_job(args['jid'])
        return rest_ok(job)
    
    @requires('authenticated')
    async def ulist_get(self, request):
        """Gets the whole ulist, with 1024 limit."""
        label2 = 'ulist_get'
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['email','ulist'], label='REST', label2=label2)
        email,ulist = itemgetter('email','ulist')(args)
        res = await self.db.select('ulists', filters={'email':email,'ulist':ulist}, sort='lseq', limit=1024)
        return rest_ok(res)
    
    @requires('authenticated')
    async def ulist_insert(self, request):
        """Gets the whole ulist, with 1024 limit."""
        label2 = 'ulist_insert'
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['email','ulist','short'], label='REST', label2=label2)
        kwargs = get_optional_args(data, ['idx','payload'])
        email,ulist,short = itemgetter('email','ulist','short')(args)
        d = {'email':email, 'ulist':ulist, 'short':short}
        if kwargs['idx'] is not None: d['idx'] = kwargs['idx']
        if kwargs['payload'] is not None: d['payload'] = kwargs['payload']
        l = LexIdx()
        seq = await self.db.query(f"select max(lseq) from ulists where email='{email}' and ulist='{ulist}'")
        seq2 = l.next(seq[0]['max'] or 0)
        d['lseq'] = seq2
        await self.db.insert('ulists', d)
        return rest_ok()

    @requires('authenticated')
    async def ulist_delete(self, request):
        """Gets the whole ulist, with 1024 limit."""
        label2 = 'ulist_delete'
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['email','ulist','lseq'], label='REST', label2=label2)
        email,ulist,lseq = itemgetter('email','ulist','lseq')(args)
        await self.db.query(f"delete from ulists where email={email} and ulist='{ulist}' and lseq='{lseq}'")
        return rest_ok()

    @requires('authenticated')
    async def ulist_update(self, request):
        """Gets the whole ulist, with 1024 limit."""
        label2 = 'ulist_update'
        u = await self.auth.check_auth(request)
        data = await request.json()
        args = enforce_required_args(self.LOG, data, ['email','ulist','lseq','lseq1','lseq2'], label='REST', label2=label2)
        email,ulist,lseq = itemgetter('email','ulist','lseq')(args)
        l = LexIdx()
        b1,b2 = l.encode(args['lseq1']),l.encode(args['lseq2'])
        seq2 = l.interpolate(b1,b2)
        await self.db.query(f"update ulists set lseq={seq2} where email='{email}' and ulist='{ulist}' and lseq='{lseq}")
        return rest_ok()

    @requires('authenticated')
    async def app_route(self, request):
        """Implements all the custom app routes."""
        label2 = f'app_route {request.url.path}'
        if request.url.path not in self.routes:
            log_and_raise_rest(self.LOG, label='REST', label2=label2, msg='url path not found', status_code=400)
        d = self.routes[request.url.path]
        cb,args,kwargs = d['f'],d['args'],d['kwargs']
        u = await self.auth.check_auth(request)
        data = await request.json()
        args2 = enforce_required_args(self.LOG, data, args, label='REST', label2=label2)
        kwargs2 = get_optional_args(data, kwargs)
        self.perm(self.d, u, request.url.path, {**args2, **kwargs2})
        try:
            res = await cb(self.d, u, *[args2[e] for e in args2], **kwargs2)
        except HTTPException as exc:
            raise
        except Exception as exc:
            log_and_raise_exception(self.LOG, label='REST', label2=label2, msg=str(exc))
        return rest_ok(res)



"""
    auth.py
    Class for users authentication backend

    All REST and downstream service errors are handled through log_and_raise_service
    All Auth REST facing methods return rest_ok
        {status:'OK, [user_status:<user_status>], [auth_status:<auth_status>, auth_msg:<auth_error_msg>]}

    Internal methods:
        get_routes
        startup
        shutdown
        check_auth

    REST methods for users:
        auth_info
        check_user
        check_otp
        set_passwd
        login
        recover_passwd
        change_passwd
        register

    REST methods for admin:
        get_users
        update_user
        reset_user
        suspend_user
        delete_user
        change_org

    Links:
    https://www.starlette.io/authentication/
    https://github.com/amitripshtos/starlette-jwt

    TODO: Response for Auth errors like WRONG_PASSWD, SUSPENDED_FAILS is 200 with STATUS OK, should be HTTP Exception 401,403.

"""
from datetime import datetime, timezone
import bcrypt
from starlette.routing import Route
from dl2050utils.core import oget
from dl2050utils.com import send_otp
from dl2050utils.db import res_to_recs, rec_to_row
from dl2050utils.restutils import log_and_raise_service, enforce_required_args, rest_ok, mk_key, mk_jwt_token

# ####################################################################################################
# Defenitions and helper functions
# ####################################################################################################

MAX_FAILS = 5
MAX_REQUESTS = 1e6

# user_status
NOT_REGISTERED = 0
WAIT_SET_PASSWD = 1
SUSPENDED = 2
SUSPENDED_LICENCE = 3
SUSPENDED_FAILS = 4
SUSPENDED_REQUESTS = 5
ACTIVE = 10

def increase(d, f, reset=False):
    """ Increases field f in user dict d. Updated ts_access, and ts_login if it is a login route """
    if d is None or f not in d: return
    if d[f] is None or reset: d[f] = 0
    d[f] += 1
    d['ts_access'] = datetime.now(timezone.utc)
    if f=='logins': d['ts_login'] = datetime.now(timezone.utc)

def get_age_in_seconds(ds):
    s1 = int(datetime.now(timezone.utc).strftime("%s"))
    s2 = int(datetime.strptime(ds,'%Y-%m-%d %H:%M:%S').strftime("%s"))
    return s1-s2

def get_user_info(user):
    return {
        'uid': user['id'],
        'email': user['email'],
        'short': user['short'],
        'name': user['name'],
        'status': user['ustatus'],
        'role': user['role'],
        'org': user['org'],
        'dep': user['org']
    }

def check_passwd_strengh(passwd):
    if type(passwd)!=str: return False
    if len(passwd)<6: return False
    return True

# ####################################################################################################
# Auth class
# ####################################################################################################

class Auth():
    def __init__(self, cfg, LOG, NOTIFY, db):
        self.cfg,self.LOG,self.NOTIFY,self.db = cfg,LOG,NOTIFY,db

    # ####################################################################################################
    # Helper methods
    # ####################################################################################################

    def ret(self, auth_status=None, user_status=None, payload=None):
        """ Return helper to combine auth_status with user_status and payload """
        d = {}
        if auth_status:
            d['auth_status'] = auth_status
        if user_status: d['user_status'] = user_status
        if type(payload)==dict: d = {**d, **payload}
        return rest_ok(d)

    async def db_update(self, tbl, kc, d, prefix=''):
        res = await self.db.update(tbl, kc, d)
        if res:
            log_and_raise_service(self.LOG, label='AUTH', label2=prefix, msg='DB update error')

    async def enforce_find_user(self, d, label2=''):
        user = await self.db.select_one('users', d)
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='User not found', status_code=401)
        if user['ustatus']!=ACTIVE:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='User not active', status_code=403)
        return user

    async def enforce_admin_perm(self, request, label2='', requires_su=False):
        user = await self.enforce_find_user({'id':request.user.display_name}, label2=label2)
        role,org = user['role'],user['org']
        roles = ['su'] if requires_su else ['su', 'admin']
        if user['ustatus']!=ACTIVE or role not in roles:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg=f'{label2}: User is not admin', status_code=403)
        return role,org

    async def increase_fails(self, user, fail='LOGIN'):
        """ Called when the user fails LOGIN, OTP or SET PASSWORD. Increases count in DB and logs.
            If fails are mode MAX_FAILS, suspends user and return True. Othewise returns False """
        increase(user, 'fails')
        if user['fails']>MAX_FAILS:
            user['ustatus'] = SUSPENDED_FAILS
            self.LOG(4, 0, label='AUTH', label2='increase_fails', msg=f"Too many {fail} fails for user {user['email']} - USER SUSPENDED")
            await self.db_update('users', 'email', user, 'increase_fails')
            return True
        await self.db_update('users', 'email', user, 'increase_fails')
        self.LOG(3, 0, label='AUTH', label2='increase_fails', msg=f"{fail} fail for user {user['email']}")
        return False

    async def gen_otp(self, user):
        """ Generates and send OTP to the user """
        if self.otp_mode=='phone' and oget(user,['phone']) is None:
            log_and_raise_service(self.LOG, label='AUTH', label2='gen_otp', msg='Phone not defined')
        otp = mk_key(n=4)
        self.LOG(2, 0, label='AUTH', label2='gen_otp', msg={'email': user['email'], 'OTP': otp})
        err = send_otp(self.NOTIFY, self.otp_mode, self.product, user['email'], user['phone'], otp)
        if err is not None:
            log_and_raise_service(self.LOG, label='AUTH', label2='gen_otp', msg='email error when sending OTP')
        user['otp'],user['ts_passwd'],user['ustatus'] =  otp, datetime.now(timezone.utc),WAIT_SET_PASSWD
        await self.db_update('users', 'email', user, 'gen_otp')

    # ####################################################################################################
    # register_requests: called from REST for every request
    # ####################################################################################################

    async def register_requests(self, url, uid, t):
        """ Used by App to register all requests and rate limit users """
        if url in ['/api/auth/is_auth', '/api/get_meta'] or uid==-1: return
        await self.db.insert('metrics', {'ts': datetime.now(timezone.utc), 'uid':uid, 'url':url, 't':t})
        user = await self.db.select_one('users', {'id': uid})
        increase(user, 'requests')
        if user['requests'] > MAX_REQUESTS:
            user['ustatus'] = SUSPENDED_REQUESTS
        await self.db_update('users', 'email', user, 'register_requests')

    # ####################################################################################################
    # Internal methods: used by App to manage Auth service
    # ####################################################################################################

    def get_routes(self):
        """" Used by App to include Auth routes in the App routes """
        return [
            Route('/api/auth/auth_info', endpoint=self.auth_info, methods=['GET']),
            Route('/api/auth/check_user', endpoint=self.check_user, methods=['POST']),
            Route('/api/auth/login', endpoint=self.login, methods=['POST']),
            Route('/api/auth/check_otp', endpoint=self.check_otp, methods=['POST']),
            Route('/api/auth/set_passwd', endpoint=self.set_passwd, methods=['POST']),
            Route('/api/auth/recover_passwd', endpoint=self.recover_passwd, methods=['POST']),
            Route('/api/auth/change_passwd', endpoint=self.change_passwd, methods=['POST']),
            Route('/api/auth/get_users', endpoint=self.get_users, methods=['POST']),
            Route('/api/auth/register', endpoint=self.register, methods=['POST']),
            Route('/api/auth/update_user', endpoint=self.update_user, methods=['POST']),
            Route('/api/auth/reset_user', endpoint=self.reset_user, methods=['POST']),
            Route('/api/auth/suspend_user', endpoint=self.suspend_user, methods=['POST']),
            Route('/api/auth/delete_user', endpoint=self.delete_user, methods=['POST']),
            Route('/api/auth/su', endpoint=self.su, methods=['POST']),
            Route('/api/auth/change_org', endpoint=self.change_org, methods=['POST']),
        ]

    async def startup(self):
        """" Used by App to configure and startup the Auth service """
        try:
            self.product = self.cfg['app']['product']
            self.auth_secret = self.cfg['rest']['auth_secret']
            self.allow_register = oget(self.cfg, ['rest', 'allow_register'], False)
            self.strong_passwd = oget(self.cfg, ['rest', 'strong_passwd'], False)
            self.otp_timeout = oget(self.cfg, ['rest', 'otp_timeout'], False)
            self.otp_mode = oget(self.cfg, ['rest', 'otp_mode'], False)
        except Exception as e:
            self.LOG(4, 0, label='AUTH', label2='startup', msg={'error_msg': str(e)})
            return True
        self.LOG(2, 0, label='AUTH', label2='startup', msg=f'OK (connected to DB:{self.db.dbname})')
        return False

    async def shutdown(self):
        """" Used by App to shutdown the Auth service """
        self.LOG(2, 0, label='AUTH', label2='shutdown', msg='OK')
        return False
    
    # ####################################################################################################
    # Internal methods (returns dict, not JSON) 
    # ####################################################################################################

    async def check_auth(self, request):
        """
            Used by App routes to check for valid users.
            Assumes that the JWT token was correctly decripted and the user passed the first level of Authenticatio.
            Then reads the email from the payload and checks in the database a valid uid,email pair.
            Finnaly checks if the user is ACTIVE.
            If all checks return the user info as dict (not json, since it is an internal method).
            Itherwise raises 401 or 403 with the corresponding error in error_msg field.
        """
        label2 = 'check_auth'
        if not hasattr(request.user, 'display_name') or not hasattr(request.user, 'payload'):
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='Invalid JWT token', status_code=401)
        uid,payload = request.user.display_name,request.user.payload
        user = await self.db.select_one('users', {'id':uid})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='User not found', status_code=401)
        if user['email']!=payload['email']:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='User email mistatch', status_code=401)
        if user['ustatus']!=ACTIVE:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='User not active', status_code=403)
        return get_user_info(user)
    
    async def get_email_from_uid(self, uid):
        """Returns email from DB based on uid."""
        user = await self.db.select_one('users', {'id': uid})
        return user['email']

    # ####################################################################################################
    # REST methods for users: all to return json
    # ####################################################################################################

    async def auth_info(self, request):
        """Returns Auth infos to be used on front-end flow."""
        d = {
            'allow_register': self.allow_register,
            'strong_passwd': self.strong_passwd,
            'otp_timeout': self.otp_timeout,
            'otp_mode': self.otp_mode
        }
        return self.ret(payload=d)

    async def check_user(self, request):
        """
            Checks if a user if able to login, searching its status by user email given as input.
            To be used before asking for password in the login process.
            Innitiates the OTP process if the User exists and Auth is not configured for self register.
            Returns auth_status and users_status, searching by user email.
            Possible auth_status returns: 'USER_OK', 'WAIT_SET_PASSWD'.
            Possible exception messages: 'USER_NOT_REGISTERED', 'USER_SUSPENDED'
        """
        label2 = 'check_user'
        data = await request.json()
        [email] = enforce_required_args(self.LOG, data, ['email'], label='AUTH', label2=label2, as_list=True)
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg=f'USER_NOT_REGISTERED: {email}', status_code=401)
        if user['ustatus']==ACTIVE:
            increase(user, 'logins')
            await self.db_update('users', 'email', user, 'check_user')
            return self.ret(auth_status='USER_OK')
        if user['ustatus']==WAIT_SET_PASSWD:
            return self.ret(auth_status='WAIT_SET_PASSWD')
        if user['ustatus']==NOT_REGISTERED and not self.allow_register:
            await self.gen_otp(user)
            return self.ret(auth_status='WAIT_SET_PASSWD')
        if user['ustatus'] in [SUSPENDED, SUSPENDED_LICENCE, SUSPENDED_FAILS, SUSPENDED_REQUESTS]:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_SUSPENDED', status_code=403)
        log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='UNKNOWN STATUS', status_code=403)

    async def check_otp(self, request):
        """
            Checks if a user has inputed the correct OTP, using as inputs the user email and the otp.
            To be used before asking for password in the set password process.
            Checks for OTP timeout, resending the OTP if timeout period has passed.
            Checks for number of OTP failures, suspending the user if the number is over the limit.
            Returns auth_status and user status if user not in WAIT_SET_PASSWD status. 
            Possible auth_status returns: 'SET_PASSWD_DONE', 'OTP_OK'
            Possible exception messages: 'OTP_ERROR', 'USER_SUSPENDED', 'WAIT_SET_PASSWD_TIMEOUT, 'WAIT_SET_PASSWD_SUSPENDED'
        """
        label2 = 'check_otp'
        data = await request.json()
        [email,otp] = enforce_required_args(self.LOG, data, ['email','otp'], label='AUTH', label2=label2, as_list=True)
        user = await self.db.select_one('users', {'email':email})
        if user['ustatus']==ACTIVE:
            return self.ret(auth_status='SET_PASSWD_DONE', user_status=user['ustatus'])
        if user['ustatus'] in [SUSPENDED, SUSPENDED_LICENCE, SUSPENDED_FAILS, SUSPENDED_REQUESTS]:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_SUSPENDED', status_code=403)
        if otp!=user['otp']:
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"OTP error: {user['email']}")
            if await self.increase_fails(user, 'OTP'):
                log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='WAIT_SET_PASSWD_SUSPENDED', status_code=403)
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='OTP_ERROR', status_code=403)
        delta = get_age_in_seconds(user['ts_passwd'])
        if delta > self.otp_timeout:
            await self.gen_otp(user)
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"OTP expired: {user['email']} (delta={delta})")
            log_and_raise_service(self.LOG, label='AUTH', msg='WAIT_SET_PASSWD_TIMEOUT', status_code=401)
        return self.ret(auth_status='OTP_OK')

    async def set_passwd(self, request):
        """
            Set a new user password if OTP and password strengh are valid.
            Possible auth_status returns: 'SET_PASSWD_OK', 'SET_PASSWD_DONE',
            Possible exception messages: 'OTP_ERROR', 'PASSWD_NOT_STRONG', 'WAIT_SET_PASSWD_TIMEOUT,
                                         'USER_SUSPENDED', 'WAIT_SET_PASSWD_SUSPENDED'
        """
        label2 = 'set_passwd'
        data = await request.json()
        [email,otp,passwd] = enforce_required_args(self.LOG, data, ['email','otp','passwd'], label='AUTH', label2=label2, as_list=True)
        user = await self.db.select_one('users', {'email':email})
        if user['ustatus']==ACTIVE:
            return self.ret(auth_status='SET_PASSWD_DONE')
        if user['ustatus'] in [SUSPENDED, SUSPENDED_LICENCE, SUSPENDED_FAILS, SUSPENDED_REQUESTS]:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_SUSPENDED', status_code=403)
        if otp!=user['otp']:
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"OTP error: {user['email']}")
            if await self.increase_fails(user, 'OTP'):
                log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='WAIT_SET_PASSWD_SUSPENDED', status_code=403)
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='OTP_ERROR', status_code=403)
        delta = get_age_in_seconds(user['ts_passwd'])
        if delta > self.otp_timeout:
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"OTP expired: {user['email']} (delta={delta})")
            await self.gen_otp(user)
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='WAIT_SET_PASSWD_TIMEOUT', status_code=403)
        if not check_passwd_strengh(passwd):
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='PASSWD_NOT_STRONG', status_code=403)
        hashed = bcrypt.hashpw(passwd.encode(), bcrypt.gensalt())
        user['passwd'],user['ts_passwd'],user['otp'],user['ustatus'] = hashed.decode(),datetime.now(timezone.utc),'',ACTIVE
        increase(user, 'logins')
        await self.db_update('users', 'email', user, 'set_passwd')
        jwt_token = mk_jwt_token(user['id'], user['email'], self.auth_secret)
        return self.ret(auth_status='SET_PASSWD_OK', payload={'jwt_token':jwt_token, 'user_info': get_user_info(user)})

    async def login(self, request):
        """
            Inputs are the email and passwd.
            Performs the user login returning auth_status and jwt_token and user_info if successfull.
            Checks if the user is pre-registered and controls number of failures.
            Possible auth_status returns: 'AUTHORIZED', 'WAIT_SET_PASSWD'
            Possible exception messages: 'WRONG_PASSWD', 'USER_SUSPENDED', 'USER_NOT_REGISTERED'
        """
        label2 = 'login'
        data = await request.json()
        [email,passwd] = enforce_required_args(self.LOG, data, ['email', 'passwd'], label='AUTH', label2=label2, as_list=True)
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"User not registered: {email}")
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_REGISTERED', status_code=401)
        if user['ustatus']==WAIT_SET_PASSWD:
            return self.ret(auth_status='WAIT_SET_PASSWD')
        if user['ustatus']==NOT_REGISTERED and not self.allow_register:
            await self.gen_otp(user)
            return self.ret(auth_status='WAIT_SET_PASSWD')
        if user['ustatus']!=ACTIVE:
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"User not active: {user['email']}")
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_SUSPENDED', status_code=403)
        if not bcrypt.checkpw(passwd.encode(), user['passwd'].encode()):
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"WRONG PASSWORD: {user['email']}")
            if await self.increase_fails(user, 'LOGIN'):
                self.LOG(3, 0, label='AUTH', label2=label2, msg=f"Too many LOGIN FAILS: {user['email']} - USER SUSPENDED")
                log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_SUSPENDED', status_code=403)
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"LOGIN FAIL: {user['email']}")
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='WRONG_PASSWD', status_code=403)
        jwt_token = mk_jwt_token(user['id'], email, self.auth_secret)
        increase(user, 'logins', True)
        user['fails'] = 0
        await self.db_update('users', 'email', user, 'login')
        return self.ret(auth_status='AUTHORIZED', payload={'jwt_token':jwt_token, 'user_info': get_user_info(user)})

    async def recover_passwd(self, request):
        """
            Set user mode in WAIT_SET_PASSWD and send new OTP
            Possible auth_status returns: 'WAIT_SET_PASSWD'
            Possible exception messages: 'USER_NOT_ACTIVE'
        """
        label2 = 'recover_passwd'
        data = await request.json()
        [email] = enforce_required_args(self.LOG, data, ['email'], label='AUTH', label2=label2, as_list=True)
        user = await self.enforce_find_user({'email':email}, label2=label2)
        if user['ustatus']==WAIT_SET_PASSWD:
            return self.ret(auth_status='WAIT_SET_PASSWD')
        if user['ustatus']!=ACTIVE:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_ACTIVE', status_code=403)
        user['ustatus'] = WAIT_SET_PASSWD
        await self.db_update('users', 'email', user, label2)
        await self.gen_otp(user)
        return self.ret(auth_status='WAIT_SET_PASSWD')

    async def change_passwd(self, request):
        """
            Changes a logged in user password.
            User header must be valid (user must be active),
            Possible auth_status returns: 'CHANGE_PASSWD_OK', 'WAIT_SET_PASSWD'
            Possible exception messages: 'WRONG_PASSWD', 'USER_NOT_ACTIVE', 'PASSWD_NOT_STRONG', 'USER_SUSPENDED'
        """
        label2 = 'change_passwd'
        user = await self.enforce_find_user({'id':request.user.display_name}, label2=label2)
        if user['ustatus']==WAIT_SET_PASSWD:
            return self.ret(auth_status='WAIT_SET_PASSWD')
        if user['ustatus']!=ACTIVE:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_ACTIVE', status_code=403)
        data = await request.json()
        [passwd,new_passwd] = enforce_required_args(self.LOG, data, ['passwd','new_passwd'], label='AUTH', label2=label2, as_list=True)
        if not check_passwd_strengh(new_passwd):
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='PASSWD_NOT_STRONG', status_code=403)
        if not bcrypt.checkpw(passwd.encode(), user['passwd'].encode()):
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"WRONG PASSWORD: {user['email']}")
            if await self.increase_fails(user, 'LOGIN'):
                self.LOG(3, 0, label='AUTH', label2=label2, msg=f"Too many LOGIN FAILS: {user['email']} - USER SUSPENDED")
                log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_SUSPENDED', status_code=403)
            self.LOG(3, 0, label='AUTH', label2=label2, msg=f"LOGIN FAIL: {user['email']}")
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='WRONG_PASSWD', status_code=403)
        hashed = bcrypt.hashpw(new_passwd.encode(), bcrypt.gensalt())
        user['passwd'],user['otp'],user['ustatus'] = hashed.decode(),'',ACTIVE
        await self.db_update('users', 'email', user, 'change_passwd')
        return self.ret(auth_status='CHANGE_PASSWD_OK')

    async def register(self, request):
        """
            Registers a new user, either in self register mode or admin pre-register
            Possible auth_status returns: 'REGISTER_OK', 'WAIT_SET_PASSWD'
            Possible exception messages: 'USER_ALREADY_REGISTERED'
        """
        label2 = 'register'
        data = await request.json()
        [email,org] = enforce_required_args(self.LOG, data, ['email','org'], label='AUTH', label2=label2, as_list=True)
        if type(org)==dict: org = oget(org,['id'])
        user2 = {'email':email, 'role':'user', 'org':org, 'ts_insert':datetime.now(timezone.utc), 'logins':0, 'fails':0, 'requests':0}
        user = await self.db.select_one('users', {'email':email})
        # Self registration
        if self.allow_register:
            if user is not None:
                log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_ALREADY_REGISTERED', status_code=401)
            user2['ustatus'] = WAIT_SET_PASSWD
            # rec_to_row
            await self.db.insert('users', user2)
            await self.gen_otp(user)
            return self.ret(auth_status='WAIT_SET_PASSWD')
        # Registration of new user by admin or su
        await self.enforce_admin_perm(request, label2=label2, requires_su=True)
        if user is not None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_ALREADY_REGISTERED', status_code=401)
        user2['ustatus'] = NOT_REGISTERED
        # rec_to_row
        await self.db.insert('users', user2)
        return self.ret(auth_status='REGISTER_OK')

    # ####################################################################################################
    # REST methods for admin: all to return json
    # ####################################################################################################

    async def get_users(self, request):
        """Get all users, possibly filtered by filters."""
        label2 = 'get_users'
        data = await request.json()
        role,org = await self.enforce_admin_perm(request, label2=label2)
        data = await request.json()
        QARGS = ['cols','filters','sfilters','sort','ascending','offset','limit']
        kwargs = {k:data[k] for k in data if k in QARGS}
        filters =  kwargs['filters'] if 'filters' in kwargs and kwargs['filters'] is not None else []
        if role=='admin': filters.append({'col':'org', 'val':org})
        kwargs['filters'] = filters
        # res = await self.db.select('users', **kwargs)
        res = await self.db.select('users', lookups=[{'col':'org', 'tbl':'orgs'}], **kwargs)
        res_to_recs(res, ['org'])
        if res is None or not 'data' in res or not 'nrows' in res:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='DB access error')
        return rest_ok({'data':res['data'], 'nrows':res['nrows']})

    async def update_user(self, request):
        """Update user data."""
        label2 = 'update_user'
        role,org = await self.enforce_admin_perm(request, label2=label2)
        data = await request.json()
        rec = rec_to_row(data)
        user = await self.db.select_one('users', {'id':rec['id']})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_REGISTERED', status_code=400)
        if role=='admin' and user['org']!=org:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='NOT_AUTHORIZED', status_code=403)
        # email update not allowed
        if 'email' in rec: del rec['email']
        await self.db_update('users', 'id', rec, label2)
        return rest_ok()

    async def reset_user(self, request):
        """Reset user to SET_NEW_PASSWORD mode and reset fails."""
        label2 = 'reset_user'
        role,org = await self.enforce_admin_perm(request, label2=label2)
        data = await request.json()
        rec = rec_to_row(data)
        user = await self.db.select_one('users', {'id':rec['id']})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_REGISTERED', status_code=400)
        if role=='admin' and user['org']!=org:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='NOT_AUTHORIZED', status_code=403)
        user['ustatus'],user['fails'],user['otp'] = 0,0,None
        await self.db_update('users', 'id', user, label2)
        return rest_ok()

    async def suspend_user(self, request):
        """Suspend a user changing ustatus to SUSPENDED or to SUSPENDED_LICENCE if rec['ustatus']==SUSPENDED_LICENCE."""
        label2 = 'suspend_user'
        role,org = await self.enforce_admin_perm(request, label2=label2)
        data = await request.json()
        rec = rec_to_row(data)
        user = await self.db.select_one('users', {'id':rec['id']})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_REGISTERED', status_code=400)
        if role=='admin' and user['org']!=org:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='NOT_AUTHORIZED', status_code=403)
        ustatus = int(rec['ustatus']) if 'ustatus' in rec and int(rec['ustatus'])==SUSPENDED_LICENCE else SUSPENDED
        user['ustatus'] = ustatus
        await self.db_update('users', 'id', user, label2)
        return rest_ok()
    
    async def delete_user(self, request):
        """Delete a user from DB."""
        label2 = 'delete_user'
        role,org = await self.enforce_admin_perm(request, label2=label2)
        data = await request.json()
        rec = rec_to_row(data)
        user = await self.db.select_one('users', {'id':rec['id']})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_REGISTERED', status_code=400)
        if role=='admin' and user['org']!=org:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='NOT_AUTHORIZED', status_code=403)
        await self.db.delete('users', 'id', rec['id'])
        return rest_ok()

    # ####################################################################################################
    # REST methods for su in debug mode
    # ####################################################################################################

    async def su(self, request):
        """Switch uid."""
        label2 = 'su'
        _,_ = await self.enforce_admin_perm(request, label2=label2, requires_su=True)
        data = await request.json()
        [uid] = enforce_required_args(self.LOG, data, ['uid'], label='AUTH', label2=label2, as_list=True)
        user = await self.enforce_find_user({'id':uid}, label2=label2)
        jwt_token = mk_jwt_token(user['id'], user['email'], self.auth_secret)
        return rest_ok({'jwt_token':jwt_token, 'user_info':get_user_info(user)})
    
    async def change_org(self, request):
        """Change su org."""
        label2 = 'change_org'
        _,_ = await self.enforce_admin_perm(request, label2=label2, requires_su=True)
        user = await self.db.select_one('users', {'id':request.user.display_name})
        if user is None:
            log_and_raise_service(self.LOG, label='AUTH', label2=label2, msg='USER_NOT_REGISTERED', status_code=400)
        data = await request.json()
        user['org'] = data['org']
        jwt_token = mk_jwt_token(user['id'], user['email'], self.auth_secret)
        _ = await self.db.update('users', 'id', user)
        return rest_ok({'jwt_token':jwt_token, 'user_info':get_user_info(user)})



"""
Google Storage (GS)
https://cloud.google.com/storage/docs/samples/storage-generate-signed-url-v4#storage_generate_signed_url_v4-python
"""

import os
import datetime
import re
import pickle
import mimetypes
from google.cloud import storage
from dl2050utils.core import oget
from dl2050utils.env import config_load
from dl2050utils.fs import json_save

class GS:
    """
    Google Cloud Storage helper class to manage buckets, files, and URLs.
    This class is a simplfied drive to interact with Google Storage, providing an abstraction
    to isolate the application level code from the underlying storage provider.
    If facilitates operations such as creating/deleting buckets, uploading/downloading files,
    generating signed URLs for blob uploads/downloads, and managing files in-memory or on disk.
    Attributes:
    default_location (str): Default location for creating GCS buckets.
    gc (storage.Client): Google Cloud storage client instance.
    """

    def __init__(self, service, default_location="europe-west1"):
        """
        Initializes the GS class with the specified Google Cloud service and location.
        Args:
            service (str): The Google Cloud service name.
            default_location (str): Default location for bucket creation. Defaults to "europe-west1".
        """
        cfg = config_load(service)
        # Create credentials file from config yml
        key_dict = oget(cfg, ["gcloud", "gs_key"])
        assert key_dict["type"] == "service_account"
        credentials_p = "./gs-keyfile.json"
        json_save(credentials_p, key_dict)
        # Set the GOOGLE_APPLICATION_CREDENTIALS env var to use the credentials file
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credentials_p
        # Connect
        self.default_location = default_location
        self.gc = storage.Client()

    # ####################################################################################################
    # Admin
    # ####################################################################################################

    def create_bucket(self, bucket_name):
        """
        Creates a new bucket in Google Cloud Storage.
        Args:
            bucket_name (str): The name of the bucket to create.
        Returns:
            int: 0 if the bucket was created successfully, 1 otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            bucket.storage_class = "STANDARD"
            new_bucket = self.gc.create_bucket(bucket, location=self.default_location)
            print(f"Bucket {new_bucket.name} created in {new_bucket.location}, storage class {new_bucket.storage_class}")
            return 0
        except Exception as exc:
            print(f"create_bucket EXCEPTION: {str(exc)}")
            return 1

    def remove_bucket(self, bucket_name, force_delete=False):
        """
        Deletes a bucket from Google Cloud Storage.
        Args:
            bucket_name (str): The name of the bucket to delete.
            force_delete (bool, optional): If True, deletes all objects in the bucket. Defaults to False.
        Returns:
            int: 0 if the bucket was deleted successfully, 1 otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            if force_delete:
                # Delete all objects in the bucket
                blobs = bucket.list_blobs()
                for blob in blobs:
                    blob.delete()
                print(f"All objects in bucket '{bucket_name}' have been deleted.")
            bucket.delete()
            print(f"Bucket '{bucket_name}' has been deleted.")
            return 0
        except Exception as exc:
            print(f"remove_bucket EXCEPTION: {str(exc)}")
            return 1

    def delete_object(self, bucket_name, blob_name):
        """
        Deletes a single object from the bucket, with exact match.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob (object) to delete.
        Returns:
            int: False if the object was deleted successfully, True otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.delete()
            print(f"Deleted object '{blob_name}' from bucket '{bucket_name}'.")
            return 0
        except Exception as exc:
            print(f"delete_object EXCEPTION: {str(exc)}")
            return 1

    def delete_objects(self, bucket_name, pattern=None, prefix=None, suffix=None):
        """
        Deletes multiple objects from a bucket matching a pattern.
        Args:
            bucket_name (str): Name of the GCS bucket.
            pattern (str, optional): Regular expression pattern to match blob names.
            prefix (str, optional): Prefix string that blob names should start with.
            suffix (str, optional): Suffix string that blob names should end with.
        Returns:
            int: The number of objects deleted.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            # Decide which blobs to list based on prefix
            blobs = bucket.list_blobs(prefix=prefix) if prefix else bucket.list_blobs()
            deleted_count = 0
            pattern_compiled = re.compile(pattern) if pattern else None
            for blob in blobs:
                blob_name = blob.name
                match = True
                if pattern_compiled:
                    if not pattern_compiled.search(blob_name):
                        match = False
                if prefix:
                    if not blob_name.startswith(prefix):
                        match = False
                if suffix:
                    if not blob_name.endswith(suffix):
                        match = False
                if match:
                    blob.delete()
                    # print(f"Deleted object '{blob_name}' from bucket '{bucket_name}'.")
                    deleted_count += 1
            print(f"Deleted {deleted_count} objects from bucket '{bucket_name}'.")
            return deleted_count
        except Exception as exc:
            print(f"delete_objects EXCEPTION: {str(exc)}")
            return 0

    def list(self, bucket_name, subdir=""):
        """
        Lists all files in a specified subdirectory of a bucket.
        Args:
            bucket_name (str): Name of the GCS bucket.
            subdir (str, optional): Subdirectory path within the bucket. Defaults to '' (root).
        Returns:
            list: List of blob names in the specified subdirectory.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blobs = bucket.list_blobs(prefix=subdir)
            file_list = [blob.name for blob in blobs if not blob.name.endswith("/")]
            print(f"Files in {bucket_name}/{subdir}: {file_list}")
            return file_list
        except Exception as exc:
            print(f"list_files EXCEPTION: {str(exc)}")
            return []

    # ###################################################################################################################
    # Memmory Download, Upload
    # ###################################################################################################################

    def upload_mem(self, bucket_name, blob_name, data, content_type="application/octet-stream", use_pickle=True):
        """
        Uploads data from memory to a specified bucket and blob.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to upload.
            data (str or bytes): Data to upload. Can be a string or bytes.
            content_type (str, optional): MIME type of the data. Defaults to 'application/octet-stream'.
            use_pickle (bool, optional): If True, serializes the data using pickle before uploading. Defaults to False.
        Returns:
            int: 0 if upload is successful, 1 otherwise.
        Examples:
            gs.upload_mem(bucket_name, blob_name, data="Hello, world!", content_type='text/plain')
            gs.upload_mem(bucket_name, blob_name, data=b'\x89PNG\r\n\x1a...', content_type='image/png')
        """
        if not use_pickle and not isinstance(data, str) and isinstance(data, bytes):
            raise TypeError('Data must be bytes or string if use_pickle is False.')
        try:
            if use_pickle:
                data = pickle.dumps(data)
            elif isinstance(data, str):
                data = data.encode("utf-8")
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.upload_from_string(data, content_type=content_type)
            print(f"Uploaded data to {bucket_name}/{blob_name}")
            return 0
        except Exception as exc:
            print(f"upload_mem EXCEPTION: {str(exc)}")
            return 1

    def download_mem(self, bucket_name, blob_name, as_string=False, encoding="utf-8", use_pickle=True):
        """
        Downloads a blob from the bucket into memory.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to download.
            as_string (bool, optional): If True, decodes the data using the specified encoding.
                                        Ignored if use_pickle is True. Defaults to False.
            encoding (str, optional): The encoding to use when decoding bytes to string. Defaults to 'utf-8'.
            use_pickle (bool, optional): If True, deserializes the data using pickle after downloading.
                                        Defaults to False.
        Returns:
            Any: The data from the blob, possibly decoded or deserialized.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            data = blob.download_as_bytes()
            if use_pickle:
                # Deserialize the data using pickle
                data = pickle.loads(data)
            elif as_string:
                # Decode the data using the specified encoding
                data = data.decode(encoding)
            # If neither use_pickle nor as_string is True, return the raw bytes
            return data
        except Exception as exc:
            print(f"download_mem EXCEPTION: {str(exc)}")
            return None

    # ###################################################################################################################
    # File Download, Upload
    # ###################################################################################################################

    def upload_file(self, bucket_name, blob_name, local_file_path, content_type="application/octet-stream"):
        """
        Uploads a local file to a specified bucket and blob.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to upload.
            local_file_path (str): Local path of the file to upload.
            content_type (str, optional): MIME type of the data. Defaults to 'application/octet-stream'.
        Returns:
            int: 0 if upload is successful, 1 otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.upload_from_filename(local_file_path, content_type=content_type)
            print(f"Uploaded {local_file_path} to {bucket_name}/{blob_name}")
            return 0
        except Exception as exc:
            print(f"upload_file EXCEPTION: {str(exc)}")
            return 1

    def download_file(self, bucket_name, blob_name, local_file_path):
        """
        Downloads a blob from the bucket to a local file.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to download.
            local_file_path (str): Local path to save the downloaded file.
        Returns:
            int: 0 if upload is successful, 1 otherwise.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.download_to_filename(local_file_path)
            print(f"Downloaded {bucket_name}/{blob_name} to {local_file_path}")
            return 0
        except Exception as exc:
            print(f"download_file EXCEPTION: {str(exc)}")
            return 1

    # ####################################################################################################
    # download_folder, upload_folder
    # ####################################################################################################

    def upload_folder(self, bucket_name, blob_name, local_folder_path):
        """
        Recursively uploads a local folder to the remote storage, preserving the directory structure.
        Args:
            bucket_name (str): Name of the remote bucket.
            blob_name (str, optional): Remote folder path within the bucket. Defaults to ''.
            local_folder_path (str): Path to the local folder to upload.
        """
        try:
            for root,dirs,files in os.walk(local_folder_path):
                for file in files:
                    # Full local file path
                    local_file_path = os.path.join(root, file)
                    # Compute relative path to maintain directory structure
                    relative_path = os.path.relpath(local_file_path, local_folder_path)
                    relative_path = relative_path.replace(os.sep, "/")  # Ensure UNIX-style path separators
                    # Construct remote file path
                    remote_file_path = os.path.join(blob_name, relative_path).replace(os.sep, "/")
                    # Upload the file
                    content_type, _ = mimetypes.guess_type(local_file_path)
                    if content_type is None: content_type = "application/octet-stream"
                    not_success = self.upload_file(bucket_name, remote_file_path, local_file_path, content_type=content_type)
                    if not_success: print(f"Failed to upload {local_file_path}")
            print(f"Uploaded folder '{local_folder_path}' to '{bucket_name}/{blob_name}'")
        except Exception as exc:
            print(f"upload_folder EXCEPTION: {str(exc)}")

    def download_folder(self, bucket_name, blob_name, local_folder_path):
        """
        Recursively downloads a folder from the remote storage to a local directory, preserving the directory structure.
        Args:
            bucket_name (str): Name of the remote bucket.
            blob_name (str): Remote folder path within the bucket.
            local_folder_path (str): Local path where the folder will be downloaded.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blobs = bucket.list_blobs(prefix=blob_name)
            for blob in blobs:
                # Skip if the blob is a directory placeholder
                if blob.name.endswith("/"): continue
                # Compute relative path to maintain directory structure
                relative_path = os.path.relpath(blob.name, blob_name)
                relative_path = relative_path.replace(os.sep, "/")
                # Construct local file path
                local_file_path = os.path.join(local_folder_path, relative_path)
                # Ensure local directories exist
                local_dir = os.path.dirname(local_file_path)
                if not os.path.exists(local_dir): os.makedirs(local_dir)
                # Download the file
                result = self.download_file(bucket_name, blob.name, local_file_path)
                if result != 0: print(f"Failed to download {blob.name}")
            print(f"Downloaded folder '{bucket_name}/{blob_name}' to '{local_folder_path}'")
        except Exception as exc:
            print(f"download_folder EXCEPTION: {str(exc)}")

    # ####################################################################################################
    # Signed urls
    # ####################################################################################################

    def upload_url(self, bucket_name, blob_name, timeout=15 * 60, size=None):
        """
        Generates a signed URL for uploading a blob.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to upload.
            timeout (int, optional): URL expiration time in seconds. Defaults to 15 minutes.
            size (int, optional): Maximum allowed size of the upload in bytes.
        Returns:
            str or None: Signed URL for uploading or None if an error occurs.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            query_parameters = (
                None if size is None else {"x-goog-content-length-range": f"0,{size}"}
            )
            url = blob.generate_signed_url(
                version="v4",
                expiration=datetime.timedelta(seconds=timeout),
                method="PUT",
                content_type="application/octet-stream",
                query_parameters=query_parameters,
            )
            return url
        except Exception as exc:
            print(f"upload_url EXCEPTION: {str(exc)}")
            return None

    def download_url(self, bucket_name, blob_name, timeout=24 * 3600):
        """
        Generates a signed URL for downloading a blob.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob to download.
            timeout (int, optional): URL expiration time in seconds. Defaults to 24 hours.
        Returns:
            str or None: Signed URL for downloading or None if an error occurs.
        """
        try:
            bucket = self.gc.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            url = blob.generate_signed_url(
                version="v4",
                expiration=datetime.timedelta(seconds=timeout),
                method="GET",
            )
            return url
        except Exception as exc:
            print(f"gcs_download_signed_url EXCEPTION: {str(exc)}")
            return None

    def urls(self, bucket_name, blob_name, timeout=24 * 3600, size=None):
        """
        Generates both upload and download signed URLs for a blob.
        Args:
            bucket_name (str): Name of the GCS bucket.
            blob_name (str): Name of the blob.
            timeout (int, optional): URL expiration time in seconds. Defaults to 24 hours.
            size (int, optional): Maximum allowed size of the upload in bytes.
        Returns:
            tuple: (upload_url, download_url)
        """
        return self.upload_url(bucket_name, blob_name, timeout=timeout, size=size),\
               self.download_url(bucket_name, blob_name, timeout=timeout)
